"""Back-end using Apache Pulsar."""

import asyncio
import functools
import logging
import os
from typing import AsyncGenerator, Optional

import pulsar  # type: ignore

from .. import broker_client_interface, log_msgs
from ..broker_client_interface import (
    ClosingFailedException,
    Message,
    MQClientException,
    Pub,
    RawQueue,
    Sub,
)
from . import utils

LOGGER = logging.getLogger("mqclient.pulsar")


class Pulsar(RawQueue):
    """Base Pulsar wrapper.

    Extends:
        RawQueue
    """

    def __init__(
        self,
        address: str,
        topic: str,
        auth_token: str,
    ) -> None:
        """Set address, topic, and client.

        Arguments:
            address {str} -- the pulsar server address, if address doesn't start with 'pulsar', append 'pulsar://'
            topic {str} -- the name of the topic
            auth_token {str} -- the (jwt) authentication token
        """
        super().__init__()
        LOGGER.info(f"Requested MQClient for topic '{topic}' @ {address}")

        self.address = address
        if not self.address.startswith("pulsar"):
            self.address = "pulsar://" + self.address
        self.topic = topic
        self.client: pulsar.Client = None
        self.auth = pulsar.AuthenticationToken(auth_token) if auth_token else None
        self._auth_token = auth_token

    async def connect(self) -> None:
        """Set up client."""
        await super().connect()
        self.client = pulsar.Client(self.address, authentication=self.auth)

    async def close(self) -> None:
        """Close client."""
        await super().close()
        if not self.client:
            raise ClosingFailedException("No client to close.")
        try:
            self.client.close()
        except Exception as e:
            # https://github.com/apache/pulsar/issues/3127
            if str(e) == "Pulsar error: AlreadyClosed":
                LOGGER.warning("Attempted to close a connection that is already closed")
                return
            raise ClosingFailedException(str(e)) from e


class PulsarPub(Pulsar, Pub):
    """Wrapper around pulsar.Producer.

    Extends:
        Pulsar
        Pub
    """

    def __init__(
        self,
        address: str,
        topic: str,
        auth_token: str,
    ) -> None:
        LOGGER.debug(f"{log_msgs.INIT_PUB} ({address}; {topic})")
        super().__init__(address, topic, auth_token)
        self.producer: pulsar.Producer = None

    async def connect(self) -> None:
        """Connect to producer."""
        LOGGER.debug(log_msgs.CONNECTING_PUB)
        await super().connect()

        # create sub so that subscription is created so messages are forwarded from topic
        # https://pulsar.apache.org/assets/images/pulsar-subscription-types-664733b68c7124129ca7d0e04dedcb96.png
        inner_sub = PulsarSub(
            self.address,
            self.topic,
            BrokerClient.SUBSCRIPTION_NAME,
            self._auth_token,
            prefetch=1,
        )
        await inner_sub.connect()
        await inner_sub.close()

        self.producer = self.client.create_producer(self.topic)
        LOGGER.debug(log_msgs.CONNECTED_PUB)

    async def close(self) -> None:
        """Close connection."""
        LOGGER.debug(log_msgs.CLOSING_PUB)
        await super().close()
        if not self.producer:
            raise ClosingFailedException("No producer to sub.")
        LOGGER.debug(log_msgs.CLOSED_PUB)

    async def send_message(
        self,
        msg: bytes,
        retries: int,
        retry_delay: float,
    ) -> None:
        """Send a message on a queue."""
        LOGGER.debug(log_msgs.SENDING_MESSAGE)
        if not self.producer:
            raise MQClientException("queue is not connected")

        def _send_msg():
            # use wrapper function so connection references can be updated by reconnects
            if not self.producer:
                raise MQClientException("queue is not connected")
            return self.producer.send(msg)

        await utils.auto_retry_call(
            func=_send_msg,
            retries=retries,
            retry_delay=retry_delay,
            close=self.close,
            connect=self.connect,
            nonretriable_conditions=None,
            logger=LOGGER,
        )
        LOGGER.debug(log_msgs.SENT_MESSAGE)


class PulsarSub(Pulsar, Sub):
    """Wrapper around pulsar.Consumer.

    Extends:
        Pulsar
        Sub
    """

    def __init__(
        self,
        address: str,
        topic: str,
        subscription_name: str,
        auth_token: str,
        prefetch: int,
    ) -> None:
        LOGGER.debug(f"{log_msgs.INIT_SUB} ({address}; {topic})")
        super().__init__(address, topic, auth_token)
        self.consumer: pulsar.Consumer = None
        self.subscription_name = subscription_name
        self.prefetch = prefetch

    async def connect(self) -> None:
        """Connect to subscriber."""
        LOGGER.debug(log_msgs.CONNECTING_SUB)
        await super().connect()

        ack_timeout: Optional[int]  # for mypy
        try:
            ack_timeout = int(os.environ["PULSAR_UNACKED_MESSAGES_TIMEOUT_SEC"])
            if ack_timeout < 10:
                ack_timeout = None
        except:  # noqa: E722
            ack_timeout = None

        self.consumer = self.client.subscribe(
            self.topic,
            self.subscription_name,
            # Neither receive with timeout nor partitioned topics can be used if the consumer queue size is zero.
            receiver_queue_size=max(self.prefetch, 1),
            unacked_messages_timeout_ms=ack_timeout,
            consumer_type=pulsar.ConsumerType.Shared,
            initial_position=pulsar.InitialPosition.Earliest,
            negative_ack_redelivery_delay_ms=0,
        )
        LOGGER.debug(log_msgs.CONNECTED_SUB)

    async def close(self) -> None:
        """Close client and redeliver any unacknowledged messages."""
        LOGGER.debug(log_msgs.CLOSING_SUB)
        if not self.consumer:
            raise ClosingFailedException("No consumer to close.")
        await asyncio.sleep(0.1)
        self.consumer.redeliver_unacknowledged_messages()
        await super().close()
        LOGGER.debug(log_msgs.CLOSED_SUB)

    @staticmethod
    def _to_message(  # type: ignore[override]  # noqa: F821 # pylint: disable=W0221
        msg: pulsar.Message,
    ) -> Optional[Message]:
        """Transform Puslar-Message to Message type."""
        id_, data = msg.message_id(), msg.data()

        if id_ is None or data is None:  # message_id may be 0; data may be b''
            return None

        # Need to serialize id? (message_id.serialize() -> bytes)
        if isinstance(id_, pulsar._pulsar.MessageId):  # pylint: disable=I1101,W0212
            return Message(id_.serialize(), data)
        # Send original data
        else:
            return Message(id_, data)

    async def get_message(
        self,
        timeout_millis: Optional[int],
        retries: int,
        retry_delay: float,
    ) -> Optional[Message]:
        """Get a single message from a queue.

        To endlessly block until a message is available, set
        `timeout_millis=None`.
        """
        LOGGER.debug(log_msgs.GETMSG_RECEIVE_MESSAGE)
        if not self.consumer:
            raise MQClientException("queue is not connected")

        def _get_msg():
            # use wrapper function so connection references can be updated by reconnects
            if not self.consumer:
                raise MQClientException("queue is not connected")
            return self.consumer.receive(timeout_millis=timeout_millis)

        try:
            pulsar_msg = await utils.auto_retry_call(
                func=_get_msg,
                retries=retries,
                retry_delay=retry_delay,
                close=None,
                connect=None,
                logger=LOGGER,
                nonretriable_conditions=lambda e: str(e) == "Pulsar error: TimeOut",
            )
        except Exception as e:
            # https://github.com/apache/pulsar/issues/3127
            # consumer timed out so there's nothing left in the tube
            if str(e) == "Pulsar error: TimeOut":
                LOGGER.debug(log_msgs.GETMSG_TIMEOUT_ERROR)
                return None
            raise
        if msg := PulsarSub._to_message(pulsar_msg):
            LOGGER.debug(f"{log_msgs.GETMSG_RECEIVED_MESSAGE} ({msg}).")
            return msg
        else:
            LOGGER.debug(log_msgs.GETMSG_NO_MESSAGE)
            return None

    async def ack_message(
        self,
        msg: Message,
        retries: int,
        retry_delay: float,
    ) -> None:
        """Ack a message from the queue."""
        LOGGER.debug(log_msgs.ACKING_MESSAGE)
        if not self.consumer:
            raise MQClientException("queue is not connected")

        if isinstance(msg.msg_id, bytes):
            pulsar_msg = pulsar.MessageId.deserialize(msg.msg_id)
        else:
            pulsar_msg = msg.msg_id

        await utils.auto_retry_call(
            func=functools.partial(
                self.consumer.acknowledge,
                pulsar_msg,
            ),
            retries=retries,
            retry_delay=retry_delay,
            close=None,
            connect=None,
            nonretriable_conditions=None,
            logger=LOGGER,
        )
        LOGGER.debug(f"{log_msgs.ACKED_MESSAGE} ({msg}).")

    async def reject_message(
        self,
        msg: Message,
        retries: int,
        retry_delay: float,
    ) -> None:
        """Reject (nack) a message from the queue."""
        LOGGER.debug(log_msgs.NACKING_MESSAGE)
        if not self.consumer:
            raise MQClientException("queue is not connected")

        if isinstance(msg.msg_id, bytes):
            pulsar_msg = pulsar.MessageId.deserialize(msg.msg_id)
        else:
            pulsar_msg = msg.msg_id

        await utils.auto_retry_call(
            func=functools.partial(
                self.consumer.negative_acknowledge,
                pulsar_msg,
            ),
            retries=retries,
            retry_delay=retry_delay,
            close=None,
            connect=None,
            nonretriable_conditions=None,
            logger=LOGGER,
        )
        LOGGER.debug(f"{log_msgs.NACKED_MESSAGE} ({msg}).")

    async def message_generator(
        self,
        timeout: int,
        propagate_error: bool,
        retries: int,
        retry_delay: float,
    ) -> AsyncGenerator[Optional[Message], None]:
        """Yield Messages.

        Generate messages with variable timeout.
        Yield `None` on `throw()`.

        Keyword Arguments:
            timeout {int} -- timeout in seconds for inactivity (default: {60})
            propagate_error {bool} -- should errors from downstream code kill the generator? (default: {True})
        """
        LOGGER.debug(log_msgs.MSGGEN_ENTERED)
        if not self.consumer:
            raise MQClientException("queue is not connected")

        msg = None
        try:
            while True:
                # get message
                LOGGER.debug(log_msgs.MSGGEN_GET_NEW_MESSAGE)
                msg = await self.get_message(
                    timeout_millis=timeout * 1000,
                    retries=retries,
                    retry_delay=retry_delay,
                )
                if msg is None:
                    LOGGER.info(log_msgs.MSGGEN_NO_MESSAGE_LOOK_BACK_IN_QUEUE)
                    break

                # yield message to consumer
                try:
                    LOGGER.debug(f"{log_msgs.MSGGEN_YIELDING_MESSAGE} [{msg}]")
                    yield msg
                # consumer throws Exception...
                except Exception as e:  # pylint: disable=W0703
                    LOGGER.debug(log_msgs.MSGGEN_DOWNSTREAM_ERROR)
                    if propagate_error:
                        LOGGER.debug(log_msgs.MSGGEN_PROPAGATING_ERROR)
                        raise
                    LOGGER.warning(
                        f"{log_msgs.MSGGEN_EXCEPTED_DOWNSTREAM_ERROR} {e}.",
                        exc_info=True,
                    )
                    yield None  # hand back to consumer
                # consumer requests again, aka next()
                else:
                    pass

        # garbage collection (or explicit generator close(), or break in consumer's loop)
        except GeneratorExit:
            LOGGER.debug(log_msgs.MSGGEN_GENERATOR_EXITING)
            LOGGER.debug(log_msgs.MSGGEN_GENERATOR_EXITED)

        # Done with generator, one way or another
        finally:
            pass


class BrokerClient(broker_client_interface.BrokerClient):
    """Pulsar Pub-Sub BrokerClient Factory.

    Extends:
        BrokerClient
    """

    NAME = "pulsar"

    # NOTE - use single shared subscription
    # (making multiple unique subscription names would create independent subscriptions)
    SUBSCRIPTION_NAME = "i3-pulsar-sub"

    @staticmethod
    async def create_pub_queue(
        address: str,
        name: str,
        auth_token: str,
    ) -> PulsarPub:
        """Create a publishing queue."""
        q = PulsarPub(  # pylint: disable=invalid-name
            address,
            name,
            auth_token,
        )
        await q.connect()
        return q

    @staticmethod
    async def create_sub_queue(
        address: str,
        name: str,
        prefetch: int,
        auth_token: str,
    ) -> PulsarSub:
        """Create a subscription queue."""
        q = PulsarSub(  # pylint: disable=invalid-name
            address,
            name,
            BrokerClient.SUBSCRIPTION_NAME,
            auth_token,
            prefetch,
        )
        await q.connect()
        return q



"""Parent class for broker_client unit tests."""

# pylint:disable=invalid-name,protected-access

import logging
from typing import Any, List, Optional
from unittest.mock import Mock

import asyncstdlib as asl
import pytest
from mqclient.broker_client_interface import BrokerClient, Message
from mqclient.config import DEFAULT_RETRIES, DEFAULT_RETRY_DELAY, DEFAULT_TIMEOUT
from mqclient.queue import Queue

from .utils import is_inst_name


class BrokerClientUnitTest:
    """Unit test suite interface for specified broker_client."""

    broker_client: BrokerClient
    con_patch = ""

    @pytest.fixture
    def mock_con(self, mocker: Any) -> Any:
        """Patch mock_con."""
        return mocker.patch(self.con_patch)

    @staticmethod
    @pytest.fixture
    def queue_name() -> str:
        """Get random queue name."""
        name = Queue.make_name()
        logging.info(f"NAME :: {name}")
        return name

    @staticmethod
    def _assert_nack_mock(mock_con: Any, called: bool, *with_args: Any) -> None:
        """Assert mock 'nack' function called (or not)."""
        raise NotImplementedError()

    @staticmethod
    def _assert_ack_mock(mock_con: Any, called: bool, *with_args: Any) -> None:
        """Assert mock 'ack' function called (or not)."""
        raise NotImplementedError()

    @staticmethod
    def _get_close_mock_fn(mock_con: Any) -> Mock:
        """Return mock 'close' function."""
        raise NotImplementedError()

    @staticmethod
    async def _enqueue_mock_messages(
        mock_con: Any, data: List[bytes], ids: List[int], append_none: bool = True
    ) -> None:
        """Place messages on the mock queue."""
        raise NotImplementedError()

    @pytest.mark.asyncio
    async def test_create_pub_queue(self, mock_con: Any, queue_name: str) -> None:
        """Test creating pub queue."""
        raise NotImplementedError()

    @pytest.mark.asyncio
    async def test_create_sub_queue(self, mock_con: Any, queue_name: str) -> None:
        """Test creating sub queue."""
        raise NotImplementedError()

    @pytest.mark.asyncio
    async def test_send_message(self, mock_con: Any, queue_name: str) -> None:
        """Test sending message."""
        raise NotImplementedError()

    @pytest.mark.asyncio
    async def test_get_message(self, mock_con: Any, queue_name: str) -> None:
        """Test getting message."""
        raise NotImplementedError()

    @pytest.mark.asyncio
    async def test_ack_message(self, mock_con: Any, queue_name: str) -> None:
        """Test acking message."""
        sub = await self.broker_client.create_sub_queue("localhost", queue_name, 1, "")

        if is_inst_name(
            self.broker_client, "rabbitmq.BrokerClient"
        ):  # HACK: manually set attr
            mock_con.return_value.is_closed = False
            sub._get_channel_by_msg = lambda *args: sub.active_channels[0]  # type: ignore[attr-defined]

        await sub.ack_message(
            Message(12, b""),
            retries=DEFAULT_RETRIES,
            retry_delay=DEFAULT_RETRY_DELAY,
        )

        self._assert_ack_mock(mock_con, True, 12)

    @pytest.mark.asyncio
    async def test_reject_message(self, mock_con: Any, queue_name: str) -> None:
        """Test rejecting message."""
        sub = await self.broker_client.create_sub_queue("localhost", queue_name, 1, "")

        if is_inst_name(
            self.broker_client, "rabbitmq.BrokerClient"
        ):  # HACK: manually set attr
            mock_con.return_value.is_closed = False
            sub._get_channel_by_msg = lambda *args: sub.active_channels[0]  # type: ignore[attr-defined]

        await sub.reject_message(
            Message(12, b""),
            retries=DEFAULT_RETRIES,
            retry_delay=DEFAULT_RETRY_DELAY,
        )

        self._assert_nack_mock(mock_con, True, 12)

    @pytest.mark.asyncio
    async def test_message_generator_00(self, mock_con: Any, queue_name: str) -> None:
        """Test message generator."""
        sub = await self.broker_client.create_sub_queue("localhost", queue_name, 1, "")
        if is_inst_name(
            self.broker_client, "rabbitmq.BrokerClient"
        ):  # HACK: manually set attr
            mock_con.return_value.is_closed = False

        num_msgs = 100

        fake_data = ["baz-{i}".encode("utf-8") for i in range(num_msgs)]
        fake_ids = [i * 10 for i in range(num_msgs)]
        await self._enqueue_mock_messages(mock_con, fake_data, fake_ids)

        msg: Optional[Message]
        async for i, msg in asl.enumerate(
            sub.message_generator(
                timeout=DEFAULT_TIMEOUT,
                propagate_error=True,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            )
        ):
            logging.debug(i)
            if i > 0:  # see if previous msg was acked
                # prev_id = (i - 1) * 10
                # would be called by Queue
                self._assert_ack_mock(mock_con, False)
            assert msg is not None
            assert msg.msg_id == fake_ids[i]
            assert msg.payload == fake_data[i]

        # last_id = (num_msgs - 1) * 10
        self._assert_ack_mock(mock_con, False)  # would be called by Queue
        # would be called by Queue
        self._get_close_mock_fn(mock_con).assert_not_called()

    @pytest.mark.asyncio
    async def test_message_generator_01(self, mock_con: Any, queue_name: str) -> None:
        """Test message generator."""
        sub = await self.broker_client.create_sub_queue("localhost", queue_name, 1, "")
        if is_inst_name(
            self.broker_client, "rabbitmq.BrokerClient"
        ):  # HACK: manually set attr
            mock_con.return_value.is_closed = False

        fake_data = [b"foo, bar", b"baz"]
        fake_ids = [12, 20]
        await self._enqueue_mock_messages(
            mock_con, fake_data, fake_ids, append_none=False
        )

        m = None
        msg: Optional[Message]
        async for i, msg in asl.enumerate(
            sub.message_generator(
                timeout=DEFAULT_TIMEOUT,
                propagate_error=True,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            )
        ):
            m = msg
            if i == 0:
                break

        assert m is not None
        assert m.msg_id == 12
        assert m.payload == b"foo, bar"
        self._assert_ack_mock(mock_con, False)  # would be called by Queue
        # would be called by Queue
        self._get_close_mock_fn(mock_con).assert_not_called()

    @pytest.mark.asyncio
    async def test_message_generator_02(self, mock_con: Any, queue_name: str) -> None:
        """Test message generator."""
        sub = await self.broker_client.create_sub_queue("localhost", queue_name, 1, "")
        if is_inst_name(
            self.broker_client, "rabbitmq.BrokerClient"
        ):  # HACK: manually set attr
            mock_con.return_value.is_closed = False

        await self._enqueue_mock_messages(mock_con, [b"foo, bar"], [12])

        m = None
        msg: Optional[Message]
        async for i, msg in asl.enumerate(
            sub.message_generator(
                timeout=DEFAULT_TIMEOUT,
                propagate_error=True,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            )
        ):
            assert i < 1
            m = msg
        assert m is not None
        assert m.msg_id == 12
        assert m.payload == b"foo, bar"
        self._assert_ack_mock(mock_con, False)  # would be called by Queue
        # would be called by Queue
        self._get_close_mock_fn(mock_con).assert_not_called()

    @pytest.mark.asyncio
    async def test_message_generator_10_upstream_error(
        self, mock_con: Any, queue_name: str
    ) -> None:
        """Failure-test message generator.

        Generator should raise Exception originating upstream (a.k.a.
        from package code).
        """
        raise NotImplementedError()

    @pytest.mark.asyncio
    async def test_message_generator_20_no_auto_ack(
        self, mock_con: Any, queue_name: str
    ) -> None:
        """Test message generator.

        Generator should not ack messages.
        """
        sub = await self.broker_client.create_sub_queue("localhost", queue_name, 1, "")
        if is_inst_name(
            self.broker_client, "rabbitmq.BrokerClient"
        ):  # HACK: manually set attr
            mock_con.return_value.is_closed = False

        fake_data = [b"baz-0", b"baz-1", b"baz-2"]
        fake_ids = [0, 1, 2]
        await self._enqueue_mock_messages(mock_con, fake_data, fake_ids)

        gen = sub.message_generator(
            timeout=DEFAULT_TIMEOUT,
            propagate_error=True,
            retries=DEFAULT_RETRIES,
            retry_delay=DEFAULT_RETRY_DELAY,
        )
        i = 0
        async for msg in gen:
            logging.debug(i)
            if i > 0:  # see if previous msg was acked
                # would be called by Queue
                self._assert_ack_mock(mock_con, False)

            assert msg is not None
            assert msg.msg_id == i
            assert msg.payload == fake_data[i]

            i += 1

    @pytest.mark.asyncio
    async def test_message_generator_30_propagate_error(
        self, mock_con: Any, queue_name: str
    ) -> None:
        """Failure-test message generator.

        Generator should raise Exception, nack, and close. Unlike in an
        integration test, nacked messages are not put back on the queue.
        """
        sub = await self.broker_client.create_sub_queue("localhost", queue_name, 1, "")
        if is_inst_name(
            self.broker_client, "rabbitmq.BrokerClient"
        ):  # HACK: manually set attr
            mock_con.return_value.is_closed = False

        fake_data = [b"baz-0", b"baz-1", b"baz-2"]
        fake_ids = [0, 1, 2]
        await self._enqueue_mock_messages(
            mock_con, fake_data, fake_ids, append_none=False
        )

        gen = sub.message_generator(
            timeout=DEFAULT_TIMEOUT,
            propagate_error=True,
            retries=DEFAULT_RETRIES,
            retry_delay=DEFAULT_RETRY_DELAY,
        )  # propagate_error=True
        i = 0
        async for msg in gen:
            logging.debug(i)
            assert i < 3
            if i > 0:  # see if previous msg was acked
                # would be called by Queue
                self._assert_ack_mock(mock_con, False)

            assert msg is not None
            assert msg.msg_id == i
            assert msg.payload == fake_data[i]

            if i == 2:
                with pytest.raises(Exception):
                    await gen.athrow(Exception)
                # would be called by Queue
                self._assert_nack_mock(mock_con, False)
                # would be called by Queue
                self._get_close_mock_fn(mock_con).assert_not_called()

            i += 1

    @pytest.mark.asyncio
    async def test_message_generator_40_suppress_error(
        self, mock_con: Any, queue_name: str
    ) -> None:
        """Failure-test message generator.

        Generator should not raise Exception. Unlike in an integration
        test, nacked messages are not put back on the queue.
        """
        sub = await self.broker_client.create_sub_queue("localhost", queue_name, 1, "")
        if is_inst_name(
            self.broker_client, "rabbitmq.BrokerClient"
        ):  # HACK: manually set attr
            mock_con.return_value.is_closed = False

        num_msgs = 11
        if num_msgs % 2 == 0:
            raise RuntimeError("`num_msgs` must be odd, so last message is nacked")

        fake_data = [f"baz-{i}".encode("utf-8") for i in range(num_msgs)]
        fake_ids = [i * 10 for i in range(num_msgs)]
        await self._enqueue_mock_messages(mock_con, fake_data, fake_ids)

        gen = sub.message_generator(
            timeout=DEFAULT_TIMEOUT,
            propagate_error=False,
            retries=DEFAULT_RETRIES,
            retry_delay=DEFAULT_RETRY_DELAY,
        )
        i = 0
        async for msg in gen:
            logging.debug(i)

            assert msg is not None
            assert msg.msg_id == i * 10
            assert msg.payload == fake_data[i]

            if i % 2 == 0:
                await gen.athrow(Exception)
                # would be called by Queue
                self._assert_nack_mock(mock_con, False)

            i += 1

        # would be called by Queue
        self._get_close_mock_fn(mock_con).assert_not_called()

    @pytest.mark.asyncio
    async def test_message_generator_50_consumer_exception_fail(
        self, mock_con: Any, queue_name: str
    ) -> None:
        """Failure-test message generator.

        Not so much a test, as an example of why QueueSubResource is
        needed.
        """
        sub = await self.broker_client.create_sub_queue("localhost", queue_name, 1, "")
        if is_inst_name(
            self.broker_client, "rabbitmq.BrokerClient"
        ):  # HACK: manually set attr
            mock_con.return_value.is_closed = False

        await self._enqueue_mock_messages(mock_con, [b"baz"], [0], append_none=False)

        excepted = False
        try:
            async for msg in sub.message_generator(
                timeout=DEFAULT_TIMEOUT,
                propagate_error=False,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            ):
                logging.debug(msg)
                raise Exception
        except Exception:
            excepted = True  # QueueSubResource would've suppressed the Exception
        assert excepted

        # would be called by Queue
        self._get_close_mock_fn(mock_con).assert_not_called()

        with pytest.raises(AssertionError):
            self._assert_nack_mock(mock_con, True, 0)

    @pytest.mark.asyncio
    async def test_queue_recv_00_consumer(self, mock_con: Any, queue_name: str) -> None:
        """Test Queue.open_sub()."""
        q = Queue(self.broker_client.NAME, address="localhost", name=queue_name)
        if is_inst_name(
            self.broker_client, "rabbitmq.BrokerClient"
        ):  # HACK: manually set attr
            mock_con.return_value.is_closed = False

        fake_data = [Message.serialize("baz")]
        await self._enqueue_mock_messages(mock_con, fake_data, [0])

        async with q.open_sub() as gen:
            async for msg in gen:
                logging.debug(msg)
                assert msg
                assert msg == "baz"

        self._get_close_mock_fn(mock_con).assert_called()
        self._assert_ack_mock(mock_con, True, 0)

    @pytest.mark.asyncio
    async def test_queue_recv_10_comsumer_exception(
        self, mock_con: Any, queue_name: str
    ) -> None:
        """Failure-test Queue.open_sub().

        When an Exception is raised in `with` block, the Queue should:
        - close (sub) on exit
        - nack the last message
        - suppress the Exception
        """
        q = Queue(self.broker_client.NAME, address="localhost", name=queue_name)
        if is_inst_name(
            self.broker_client, "rabbitmq.BrokerClient"
        ):  # HACK: manually set attr
            mock_con.return_value.is_closed = False

        fake_data = [Message.serialize("baz-0"), Message.serialize("baz-1")]
        fake_ids = [0, 1]
        await self._enqueue_mock_messages(
            mock_con, fake_data, fake_ids, append_none=False
        )

        class TestException(Exception):  # pylint: disable=C0115
            pass

        async with q.open_sub() as gen:  # suppress_errors=True
            async for i, msg in asl.enumerate(gen):
                assert i == 0
                logging.debug(msg)
                raise TestException

        self._get_close_mock_fn(mock_con).assert_called()
        self._assert_nack_mock(mock_con, True, 0)

    @pytest.mark.asyncio
    async def test_queue_recv_11_comsumer_exception(
        self, mock_con: Any, queue_name: str
    ) -> None:
        """Failure-test Queue.open_sub().

        Same as test_queue_recv_10_comsumer_exception() but with
        multiple open_sub() calls.
        """
        q = Queue(self.broker_client.NAME, address="localhost", name=queue_name)
        if is_inst_name(
            self.broker_client, "rabbitmq.BrokerClient"
        ):  # HACK: manually set attr
            mock_con.return_value.is_closed = False

        num_msgs = 12

        fake_data = [Message.serialize(f"baz-{i}") for i in range(num_msgs)]
        fake_ids = [i * 10 for i in range(num_msgs)]
        await self._enqueue_mock_messages(mock_con, fake_data, fake_ids)

        class TestException(Exception):  # pylint: disable=C0115
            pass

        async with q.open_sub() as gen:  # suppress_errors=True
            async for msg in gen:
                logging.debug(msg)
                raise TestException

        self._get_close_mock_fn(mock_con).assert_called()
        self._assert_nack_mock(mock_con, True, 0)

        logging.info("Round 2")

        # continue where we left off
        async with q.open_sub() as gen:  # suppress_errors=True
            # self._get_mock_ack(mock_con).assert_not_called()
            async for i, msg in asl.enumerate(gen, start=1):
                logging.debug(f"{i} :: {msg}")
                if i > 1:  # see if previous msg was acked
                    prev_id = (i - 1) * 10
                    self._assert_nack_mock(mock_con, True, prev_id)

            last_id = (num_msgs - 1) * 10
            self._assert_nack_mock(mock_con, True, last_id)

        self._get_close_mock_fn(mock_con).assert_called()

    @pytest.mark.asyncio
    async def test_queue_recv_12_comsumer_exception(
        self, mock_con: Any, queue_name: str
    ) -> None:
        """Failure-test Queue.open_sub().

        Same as test_queue_recv_11_comsumer_exception() but with error
        propagation.
        """
        q = Queue(self.broker_client.NAME, address="localhost", name=queue_name)
        if is_inst_name(
            self.broker_client, "rabbitmq.BrokerClient"
        ):  # HACK: manually set attr
            mock_con.return_value.is_closed = False

        num_msgs = 12

        fake_data = [Message.serialize(f"baz-{i}") for i in range(num_msgs)]
        fake_ids = [i * 10 for i in range(num_msgs)]
        await self._enqueue_mock_messages(mock_con, fake_data, fake_ids)

        class TestException(Exception):  # pylint: disable=C0115
            pass

        with pytest.raises(TestException):
            q.except_errors = False
            async with q.open_sub() as gen:
                async for msg in gen:
                    logging.debug(msg)
                    raise TestException

        self._get_close_mock_fn(mock_con).assert_called()
        self._assert_nack_mock(mock_con, True, 0)

        logging.info("Round 2")

        # HACK:- rabbitmq deletes its connection (mock_con) when close()
        # is called, so we need to re-enqueue messages to avoid getting
        # the entire original list.
        # ***Note***: this hack isn't needed in non-mocking tests, see
        # integrate_queue.py integration tests #60+.
        if is_inst_name(q._broker_client, "rabbitmq.BrokerClient"):
            await self._enqueue_mock_messages(mock_con, fake_data[1:], fake_ids[1:])

        # continue where we left off
        q.except_errors = False
        async with q.open_sub() as gen:
            self._assert_ack_mock(mock_con, False)
            async for i, msg in asl.enumerate(gen, start=1):
                logging.debug(f"{i} :: {msg}")
                if i > 1:  # see if previous msg was acked
                    prev_id = (i - 1) * 10
                    self._assert_ack_mock(mock_con, True, prev_id)

            last_id = (num_msgs - 1) * 10
            self._assert_nack_mock(mock_con, True, last_id)

        self._get_close_mock_fn(mock_con).assert_called()



"""Unit Tests for Pulsar BrokerClient."""

from typing import Any, List

import pytest
from mqclient import broker_client_manager
from mqclient.broker_client_interface import Message
from mqclient.config import (
    DEFAULT_RETRIES,
    DEFAULT_RETRY_DELAY,
    DEFAULT_TIMEOUT,
    DEFAULT_TIMEOUT_MILLIS,
)

from ...abstract_broker_client_tests.unit_tests import BrokerClientUnitTest


class TestUnitApachePulsar(BrokerClientUnitTest):
    """Unit test suite interface for Apache Pulsar broker_client."""

    broker_client = broker_client_manager.get_broker_client("pulsar")
    con_patch = "pulsar.Client"

    @staticmethod
    def _assert_nack_mock(mock_con: Any, called: bool, *with_args: Any) -> None:
        """Assert mock 'nack' function called (or not)."""
        if called:
            mock_con.return_value.subscribe.return_value.negative_acknowledge.assert_called_with(
                *with_args,
            )
        else:
            mock_con.return_value.subscribe.return_value.negative_acknowledge.assert_not_called()

    @staticmethod
    def _assert_ack_mock(mock_con: Any, called: bool, *with_args: Any) -> None:
        """Assert mock 'ack' function called (or not)."""
        if called:
            mock_con.return_value.subscribe.return_value.acknowledge.assert_called_with(
                *with_args,
            )
        else:
            mock_con.return_value.subscribe.return_value.acknowledge.assert_not_called()

    @staticmethod
    def _get_close_mock_fn(mock_con: Any) -> Any:
        """Return mock 'close' function call."""
        return mock_con.return_value.close

    @staticmethod
    async def _enqueue_mock_messages(
        mock_con: Any, data: List[bytes], ids: List[int], append_none: bool = True
    ) -> None:
        """Place messages on the mock queue."""
        if append_none:
            data += [None]  # type: ignore
            ids += [None]  # type: ignore
        mock_con.return_value.subscribe.return_value.receive.return_value.data.side_effect = (
            data
        )
        mock_con.return_value.subscribe.return_value.receive.return_value.message_id.side_effect = (
            ids
        )

    @pytest.mark.asyncio
    async def test_create_pub_queue(self, mock_con: Any, queue_name: str) -> None:
        """Test creating pub queue."""
        pub = await self.broker_client.create_pub_queue("localhost", queue_name, "")
        assert pub.topic == queue_name  # type: ignore
        mock_con.return_value.create_producer.assert_called()

    @pytest.mark.asyncio
    async def test_create_sub_queue(self, mock_con: Any, queue_name: str) -> None:
        """Test creating sub queue."""
        sub = await self.broker_client.create_sub_queue(
            "localhost", queue_name, 213, ""
        )
        assert sub.topic == queue_name  # type: ignore
        assert sub.prefetch == 213  # type: ignore
        mock_con.return_value.subscribe.assert_called()

    @pytest.mark.asyncio
    async def test_send_message(self, mock_con: Any, queue_name: str) -> None:
        """Test sending message."""
        pub = await self.broker_client.create_pub_queue("localhost", queue_name, "")
        await pub.send_message(
            b"foo, bar, baz",
            retries=DEFAULT_RETRIES,
            retry_delay=DEFAULT_RETRY_DELAY,
        )
        mock_con.return_value.create_producer.return_value.send.assert_called_with(
            b"foo, bar, baz"
        )

    @pytest.mark.asyncio
    async def test_get_message(self, mock_con: Any, queue_name: str) -> None:
        """Test getting message."""
        sub = await self.broker_client.create_sub_queue("localhost", queue_name, 1, "")
        mock_con.return_value.subscribe.return_value.receive.return_value.data.return_value = Message.serialize(
            "foo, bar"
        )
        mock_con.return_value.subscribe.return_value.receive.return_value.message_id.return_value = (
            12
        )
        m = await sub.get_message(
            timeout_millis=DEFAULT_TIMEOUT_MILLIS,
            retries=DEFAULT_RETRIES,
            retry_delay=DEFAULT_RETRY_DELAY,
        )

        assert m is not None
        assert m.msg_id == 12
        assert m.data == "foo, bar"

    @pytest.mark.asyncio
    async def test_message_generator_10_upstream_error(
        self, mock_con: Any, queue_name: str
    ) -> None:
        """Failure-test message generator.

        Generator should raise Exception originating upstream (a.k.a.
        from pulsar-package code).
        """
        sub = await self.broker_client.create_sub_queue("localhost", queue_name, 1, "")

        retries = 2  # >= 0

        class _MyException(Exception):
            pass

        mock_con.return_value.subscribe.return_value.receive.side_effect = (
            _MyException()
        )
        with pytest.raises(_MyException):
            async for m in sub.message_generator(
                timeout=DEFAULT_TIMEOUT,
                propagate_error=True,
                retries=retries,
                retry_delay=DEFAULT_RETRY_DELAY,
            ):
                pass
        # would be called by Queue one more time
        assert self._get_close_mock_fn(mock_con).call_count == 0

        # reset for next call
        self._get_close_mock_fn(mock_con).reset_mock()

        # `propagate_error` attribute has no affect (b/c it deals w/ *downstream* errors)
        mock_con.return_value.subscribe.return_value.receive.side_effect = (
            _MyException()
        )
        with pytest.raises(_MyException):
            async for m in sub.message_generator(
                timeout=DEFAULT_TIMEOUT,
                propagate_error=False,
                retries=retries,
                retry_delay=DEFAULT_RETRY_DELAY,
            ):
                pass
        # would be called by Queue one more time
        assert self._get_close_mock_fn(mock_con).call_count == 0
